#include <stdio.h>
#include "message.h"

/* static void DdReassignLevels(DdManager *manager);
 */
#define dist_f(x,N) (x*x*(N-1-x))

#define DD_BLOCKSIFT_MINDOWN 0.01
#define DD_BLOCKSIFT_MINUP   0.02
#define DD_BLOCKSIFT_MINBLOCK 0.1

/* #define dist_f(x,N) 1 */
int machine_number(int index,int vars,int nodes)
{
  int res,normal=0,i;
  for(i=0;i<vars;i++)
    {
      normal+=dist_f(i,vars);
      if (i==index) res=normal;
    }
  res=res*nodes/(normal+1);
  return(res);
}

int InitProcesses(DdManager *manager,int cache_size,int cache_size_max)
{
  int info,i;
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  for(i=0;i<pi->Nnode;i++)
    pvm_spawn(CUDD_SLAVE_PROGRAM, (char**)0, PvmTaskHost , pi->hostp[i].hi_name, 1, &pi->Tids[i+1]);
  /* Tids[i+1] because both process 0 and 1 are on node 0 */
  pi->Nprocess=pi->Nnode +1;

  pvm_setsbuf(pi->DownBuf); /* not needed */
  pvm_pkint(&pi->Nprocess,1,1);
  pvm_pkint(pi->Tids,pi->Nprocess,1);
  pvm_pkint(&pi->Encoding,1,1);

  PVM_pkDdNodePar(manager->one);
  pvm_pkint(&cache_size,1,1);
  pvm_pkint(&cache_size_max,1,1);

  info=pvm_mcast(pi->Tids+1, pi->Nprocess-1, DD_DOWN_MESSAGE);
  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);

  return(1);
}

int KillProcesses(DdManager *manager)
{
  int allocated;
  int i;
  ProcessInfo *pi;
  int type=CUDD_END;
  pi=manager->ProcessInfo;
  pvm_setsbuf(pi->DownBuf); /* not needed */
  pvm_pkint(&type,1,1);
  pvm_mcast(pi->Tids+1, pi->Nprocess-1, DD_DOWN_MESSAGE);
  pvm_freebuf(pi->DownBuf);
  /* pi->DownBuf=pvm_mkbuf(pi->Encoding);
     no need to make buffer again */ 
  return(1);
}

int DdMessageSendMaster2(DdManager *manager,int op,DdNodePar f,DdNodePar g)
{
  /* using index=level */
  DdMessageSend2(manager,op,ddMin(f.index,g.index),0,f,g,0);
  manager->ProcessInfo->flush_count++;
  flush_buffer_down(manager);

  return(1); /* NO ERROR HANDELING */
}

int DdMessageSendMaster3(DdManager *manager,int op,DdNodePar f,DdNodePar g,DdNodePar h)
{
  /* using index=level */
  int index;
  index=ddMin(f.index,g.index);
  index=ddMin(index,h.index);
  DdMessageSend3(manager,op,index,0,f,g,h,0);
  manager->ProcessInfo->flush_count++;
  flush_buffer_down(manager);
  return(1); /* NO ERROR HANDELING */
}

int DdInitMaster(DdManager *manager)
{
  ProcessInfo *pi;
  int narch;

  pi = (ProcessInfo *) ALLOC(ProcessInfo,1);
  pi->Mytid = pvm_mytid();
  pvm_setopt(PvmAutoErr, 1);
  pvm_catchout(stdout);
  pvm_config(&(pi->Nnode),&narch,&(pi->hostp));
  pvm_setopt(PvmRoute, PvmDontRoute);

#ifdef USE_PGTRACE
  pg_startadmin("", pi->hostp[0].hi_name, MAX_PROC);
  pg_tids("y");
#endif

  pi->Nprocess=1;          /* will increase as processes are initialised */
  pi->Tids[0]=pi->Mytid;
  manager->Mypos=0;
  pi->flush_count=0;
  
  if(narch==1)
    pi->Encoding=PvmDataRaw;
  else
    pi->Encoding=PvmDataDefault;
  pi->down_flag=FALSE;
  pi->DownCount=0;
  pi->DownCountSend=DD_DOWN_BUFFER_SIZE;
  pi->DownBuf=pvm_mkbuf(pi->Encoding);

  /* No up info as master only sends down */

  pi->NextRedistribution=DD_NODES_BETWEEN_REDISTRUBUTION;
#ifdef DD_COMM_MONITOR
  pi->DownMessages=0;
  pi->UpMessages=0;
#endif

  manager->ProcessInfo=pi;
  manager->memused += sizeof(ProcessInfo);  
  return(1); /* NO ERROR HANDELING */
}
DdNodePar DdMessageRecieveMaster(DdManager *manager)
{
  int op,to,flush_count,i;
  DdCallTreeNode caller;
  DdNodePar res;
  int n;

  pvm_recv(-1,DD_UP_MESSAGE); 
  pvm_upkint(&op,1,1);
  assert(op==CUDD_BDD_REPLY);
  pvm_upkint(&n,1,1);
  assert(n==1);
  PVM_upkDdNodePar(res);
  PVM_upkDdCallTreeNode(caller);
  assert(caller==0);

  pvm_upkint(&op,1,1);
  assert(op==CUDD_FLUSH_UP);
  pvm_upkint(&flush_count,1,1);
  assert(flush_count==manager->ProcessInfo->flush_count);
  pvm_upkint(manager->allocatedPerLevel,manager->size,1);

  manager->allocated=0;
  for(i=0;i<manager->size;i++)
    manager->allocated+=manager->allocatedPerLevel[i];

  if(manager->allocated>manager->ProcessInfo->NextRedistribution && manager->ProcessInfo->Nprocess>2)
    {
      DdReassignLevels(manager);
    }
  return(res); /* NO ERROR HANDELING */  
}

DdNodePar DdAssignNewVar(DdManager *manager)
{
  int op=CUDD_NEW_VAR;
  ProcessInfo *pi;
  DdNodePar res;
  int index,to;
  pi=manager->ProcessInfo;
  index=manager->size;
  
  pvm_setsbuf(pi->DownBuf); /* not needed */
  pvm_pkint(&op,1,1);
  /* before first redistribution all levels are given to first processor
     after that they must be given to the last */
  if(manager->allocated<DD_NODES_BETWEEN_REDISTRUBUTION)
    to=1;
  else
    to=pi->Nprocess-1;
   
  manager->Local[index]=to;
  pvm_pkint(&to,1,1);
  pvm_pkint(&index,1,1);
  pvm_send(pi->Tids[1], DD_DOWN_MESSAGE);
  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);
  pvm_recv(-1,-1);
  PVM_upkDdNodePar(res);

  return(res);
} 

/* this assumes slaves know reorder method and sends the appropriate
   information */
int cuddReorderUpdateTableSizes(DdManager *manager)
{
  int i,index,j;
  ProcessInfo *pi;
  DdHalfWord size;
  DdNodeMore **table;
  pi=manager->ProcessInfo;
  /* collect TableSizes */
  for(i=1;i<manager->ProcessInfo->Nprocess;i++)
    {
      pvm_recv(-1,0);
      pvm_upkint(&index,1,1);
      while(index!=CUDD_MAXINDEX)
	{
	  PVM_upkDdHalfWord(&size,1,1);
	  manager->memory[index].TableSize=size;
	  table=manager->memory[index].PageTable2=ALLOC(DdNodeMore *,size);
	  /* recovering from out of memory is hard here, so donet try */ 
	  assert(manager->memory[index].PageTable2!=NULL);
	  for(j=0;j<size;j++)
	    table[j]=NULL;
	  pvm_upkint(&index,1,1);
	}
    }

  /* distribute TableSizes */
  pvm_setsbuf(pi->DownBuf); /* not needed */
  for(index=0;index<manager->size;index++)
    {
      PVM_pkDdHalfWord(&manager->memory[index].TableSize,1,1);
    }
  pvm_mcast(pi->Tids+1, pi->Nprocess-1, 0);
  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);
}

int cuddReorderInitaite(DdManager *manager)
{
  int op=CUDD_REORDER;
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  pvm_setsbuf(pi->DownBuf); /* not needed */
  pvm_pkint(&op,1,1);
  pvm_mcast(pi->Tids+1, pi->Nprocess-1, DD_DOWN_MESSAGE);
  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);
}

int cuddDistributeLevels(DdManager *manager)
{
  /* implementing block restricted sifting */
  int i,j,value;
  float down;
  float up;
  int Profile[manager->size+1];
  int divide[manager->size];
  int block[manager->size];
  int block_size,min_block_size;
  int start,end;
  int ctr;
  int op;

  ProcessInfo *pi;
  pi=manager->ProcessInfo;
  /* collect Profile */
  for(i=1;i<=manager->size;i++)
    Profile[i]=0;
  for(i=1;i<manager->ProcessInfo->Nprocess;i++)
    {
      pvm_recv(-1,0);
      for(j=1;j<=manager->size;j++)
	{
	  pvm_upkint(&value,1,1);
	  Profile[j]+=value;
	}
    }
  /* find possible divisions */
  for(j=2;j<manager->size;j++)
    {
      down=(Profile[j-1]-Profile[j])/(float)Profile[j];
      up=(Profile[j+1]-Profile[j])/(float)Profile[j];
      if(down>DD_BLOCKSIFT_MINDOWN && up>DD_BLOCKSIFT_MINUP)
	divide[j]=TRUE;
      else
	divide[j]=FALSE;
    }
  
  /* discard divisions making too small blocks */
  block_size=1; /* index 0 by default in block 1 */
  min_block_size=DD_BLOCKSIFT_MINBLOCK*manager->size;
  for(j=2;j<manager->size;j++)
    {
      if(!divide[j])
	block_size++;
      else 
	{
	  if(block_size>=min_block_size)
	    block_size=0;
	  else
	    divide[j]=FALSE;
	}
    }
  
  /* distribute the blocks (loadbalancing to be considered) */

  pvm_setsbuf(pi->DownBuf); /* not needed */
  pvm_pkint(Profile+1,manager->size,1);
   
  start=0;
  ctr=0;

  for(j=2;j<manager->size;j++)
    {
      if(divide[j])    
	{
	  end=j-1;
	  for(i=start;i<=end;i++) {
	    manager->Local[i]=manager->Local[start];
	    block[i]=ctr;
	  }
	  start=end+1;
	  ctr++;
	}
    }
  
  end=manager->size-1;
  block[end]=ctr;
  for(i=start;i<=end;i++){
    manager->Local[i]=manager->Local[end];
    block[i]=ctr;
  }
  
  for(i=0;i<manager->size;i++)
    printf("index:%2d block:%d owner:%d\n",i,block[i],manager->Local[i]);
  pvm_pkint(block,manager->size,1);
  op=CUDD_REDISTRIBUTE_LEVELS;
  pvm_pkint(&op,1,1);
  pvm_pkint(manager->Local,manager->size,1);
  pvm_send(pi->Tids[1],0);
  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);

  cuddSendExternRef(manager);

  return(1); 
}

int cuddDumpDotMaster(DdManager *manager,char **inames,char **onames,char file_name[])
{
  int op=CUDD_DUMP_DOT;
  int retval;
  FILE *fp;
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  fp=fopen(file_name,"w");
  if(fp==NULL)
    retval=0;
  else
    {
      retval = cuddDumpDotMaster2(manager,inames,onames,fp);
      fclose(fp);
    }
  if(retval==0)
    return(0);
  else
    { 
      pvm_setsbuf(pi->DownBuf);
      pvm_pkint(&op,1,1);      
      pvm_pkstr(file_name);

      pvm_send(pi->Tids[1],0);
      pvm_freebuf(pi->DownBuf);
      pi->DownBuf=pvm_mkbuf(pi->Encoding);
    }
  pvm_recv(-1,0);
  pvm_upkint(&retval,1,1);
  if(retval==0)
    return(0);
  else
    {
      fp=fopen(file_name,"a");
      if(fp==NULL)
	retval=0;
      else
	{
	  retval = fprintf(fp,"}\n");
	  fclose(fp);
	}
    }
  return(retval);
}

/* static */ 
void DdReassignLevels(DdManager *manager)
{
  int *allocPerLvl=manager->allocatedPerLevel;
  int new_owner[manager->size];
  int i,owner,last_owner=0,last_index;
  int sum;
  int op=CUDD_REDISTRIBUTE_LEVELS;
  int bufid;
  int last;
  float average;
  ProcessInfo *pi;
  pi=manager->ProcessInfo;
  
  /* note: all levels have at least one node */

  average=(float)manager->allocated/(pi->Nprocess-1);

  sum=0;
  last=1;
  for(i=0;i<manager->size-1;i++)
    {
      sum+=allocPerLvl[i];
      manager->Local[i]=last;
      if(last!=ceil(sum/average))
	last++;
      printf("level %3d at %2d\n",i,manager->Local[i]);
    }
  
  /* to protect against rounding errors this is moved out of the loop */
  manager->Local[manager->size-1]=pi->Nprocess-1;

  bufid = pvm_mkbuf(pi->Encoding);
  pvm_setsbuf(bufid);
  pvm_pkint(&op,1,1);
  pvm_pkint(manager->Local,manager->size,1);
  pvm_send(pi->Tids[1],0);
  pvm_freebuf(bufid);

  pi->NextRedistribution=manager->allocated*2;
}

  

