#include <stdio.h>
#include "message.h"

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int DdRetrieveIndex(DdManager *manager, int index,int from);
static int DdRecieveIndex(DdManager *manager, int index);
static int DdSendIndex(DdManager *manager, int index,int to);
static int DdMessageSendUp(DdManager *manager);


int cuddReorderRedistributeLevelsSlave(DdManager *manager)
{
  int op;
  pvm_recv(manager->ProcessInfo->Tids[manager->Mypos-1],0);

  pvm_upkint(manager->Profile+1,manager->size,1);
  manager->ReorderBlocks=ALLOC(int,manager->size);
  pvm_upkint(manager->ReorderBlocks,manager->size,1);
  if(manager->Mypos!=manager->ProcessInfo->Nprocess-1)
    {
      pvm_setsbuf(manager->ProcessInfo->DownBuf);
      pvm_pkint(manager->Profile+1,manager->size,1);
      pvm_pkint(manager->ReorderBlocks,manager->size,1);
    }
  pvm_upkint(&op,1,1);
  assert(op==CUDD_REDISTRIBUTE_LEVELS);
  cuddRedistributeLevelsSlave(manager);
  return(1);
}

int cuddRedistributeLevelsSlave(DdManager *manager)
{
  int i,index;
  int to,op;
  int newOwner[manager->size];
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  pvm_upkint(newOwner,manager->size,1);
  if(manager->Mypos!=pi->Nprocess-1)
    { /* pass the array down */
      pvm_setsbuf(pi->DownBuf);
      op=CUDD_REDISTRIBUTE_LEVELS;
      pvm_pkint(&op,1,1);
      pvm_pkint(newOwner,manager->size,1);
      pvm_send(pi->Tids[manager->Mypos+1],0);
      pvm_freebuf(pi->DownBuf);
      pi->DownBuf=pvm_mkbuf(pi->Encoding);
    }	

  for(i=0;i<manager->size;i++)
    {
      if(cuddLocal(manager,i))
	{ 
	  if(newOwner[i]!=manager->Mypos)
	    { /* new owner will retrieve the index */
	      pvm_recv(pi->Tids[newOwner[i]],0);
	      pvm_upkint(&op,1,1);
	      assert(op==CUDD_SEND_INDEX);
	      pvm_upkint(&index,1,1);
	      assert(index==i);
	      pvm_upkint(&to,1,1);
	      assert(to==newOwner[i]);
	      DdSendIndex(manager,index,to);
	    }
	  /* else newOwner[i]==manager->Mypos and do nothing */
	}
      else if(newOwner[i]==manager->Mypos)
	{ /* retrieve index */
	  DdRetrieveIndex(manager,i,DdFindOwner(manager,i));	  
	}
      else
	cuddSetOwner(manager,i,newOwner[i]);
    }

  return(1);
}

int cuddSendProfile(DdManager *manager)
{
  int bufid;
  /* send own tablesizes */
  bufid = pvm_mkbuf(manager->ProcessInfo->Encoding);
  pvm_setsbuf(bufid);
  /* use a new buffer to avoid leaving junk in up_buf or down_buf */

  pvm_pkint(manager->Profile+1,manager->size,1);
  pvm_send(manager->ProcessInfo->Parent,0);
  pvm_freebuf(bufid);

  return(1);
}
int cuddReorderUpdateTableSizesSlave(DdManager *manager)
{
  int bufid;
  int index;
  DdHalfWord size;
  DdNodeMore **table;
  int i;

  /* send own tablesizes */
  bufid = pvm_mkbuf(manager->ProcessInfo->Encoding);
  pvm_setsbuf(bufid);
  /* use a new buffer to avoid leaving junk in up_buf or down_buf */

  for(index=0;index<manager->size;index++)
    if(cuddLocal(manager,index))
      {
	pvm_pkint(&index,1,1);
	PVM_pkDdHalfWord(&manager->memory[index].TableSize,1,1);
      }
  index=CUDD_MAXINDEX;
  pvm_pkint(&index,1,1);
  pvm_send(manager->ProcessInfo->Parent,0);
  pvm_freebuf(bufid);

  /* recieve tablesizes of other machines and reserve space for pagetables */
  pvm_recv(manager->ProcessInfo->Parent,0);
  for(index=0;index<manager->size;index++)
    {
      PVM_upkDdHalfWord(&size,1,1);
      if(!cuddLocal(manager,index))
	{
	  manager->memory[index].TableSize=size;
	  table=manager->memory[index].PageTable2=ALLOC(DdNodeMore *,size);
	  if(manager->memory[index].PageTable2==NULL)
	    return(0);
	  for(i=0;i<size;i++)
	    table[i]=NULL;
	}
      else
	assert(size==manager->memory[index].TableSize);
    }
  return(1);
}

int cuddGetExternRef(DdManager *manager) 
{
  int index,page_number,number,i;
  DdNodeMore *ref_page;

  pvm_recv(manager->ProcessInfo->Tids[manager->Mypos-1],0);

  pvm_upkint(&index,1,1);
  while(index<CUDD_MAXINDEX)
    {
      pvm_upkint(&page_number,1,1);
      while(page_number!=NULL_PAGE)
	{
	  ref_page=manager->memory[index].PageTable2[page_number];
	  if(ref_page==NULL)
	    { /* references going past this process only */
	      /* hashing should be considered */
	      ref_page=manager->memory[index].PageTable2[page_number] =
		ALLOC(DdNodeMore,DD_MEM_CHUNK)-1;
	      if(ref_page==NULL)
		return(0);
	      manager->memused+=sizeof(DdNodeMore)*DD_MEM_CHUNK;
	      for(i=1;i<=DD_MEM_CHUNK;i++)
		ref_page[i].Reorder.Ref=0;
	    }
	  
	  pvm_upkint(&number,1,1);
	  while(number!=NULL_NUMBER)
	    {
	      ref_page[number].Reorder.Ref=DD_EXTERN_REF;
	      pvm_upkint(&number,1,1);
	    }
	  pvm_upkint(&page_number,1,1);
	}

      pvm_upkint(&index,1,1);
    }
  return(1);
}

int cuddExternDistributeNewLocations(DdManager *manager,DdMemory *old_mems) 
{
  pvm_send(manager->ProcessInfo->Tids[manager->Mypos-1],0);
  pvm_freebuf(manager->ProcessInfo->UpBuf);
  manager->ProcessInfo->UpBuf=pvm_mkbuf(manager->ProcessInfo->Encoding);
  return(1); /* NO ERROR HANDELING */
}
      

DdHalfWord f_index[DD_BUFFER_SIZE];
DdHalfWord f_page[DD_BUFFER_SIZE];
DdSignedHalfWord f_number[DD_BUFFER_SIZE];
DdHalfWord g_index[DD_BUFFER_SIZE];
DdHalfWord g_page[DD_BUFFER_SIZE];
DdSignedHalfWord g_number[DD_BUFFER_SIZE];
DdCallTreeNode caller[DD_BUFFER_SIZE];


int DdReadHandleOperation(DdManager *manager,int from)
{
  int info;
  int op;
  DdNodePar f,g,h;
  int to;
  int i,n;
  int done=FALSE;
  int index;
  int flush_count;
  char file_name[100];
  int new_from;
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  info=pvm_upkint(&op,1,1);
  assert(info==0);
  switch(op)
    {
    case CUDD_END: 
      DdWorkQueuePush2(manager,op,0,0,CUDD_ERROR,CUDD_ERROR,0);
      /* note only op and manager needs well defined values */
      return(0);
      break; 

    case CUDD_BDD_REPLY:
      pvm_upkint(&n,1,1);
      pvm_upkushort(f_index, n,1);
      pvm_upkushort(f_page,  n,1);
      pvm_upkshort (f_number,n,1);
      pvm_upkint   (caller,  n,1);
      
      for(i=0;i<n;i++)
	{
	  f.index=f_index[i];
	  f.address.page=f_page[i];
	  f.address.number=f_number[i];
	  DdCallTreeUpdateNode(manager,caller[i],f);
	}
      break;
    case CUDD_BDD_EXISTS:
      op=CUDD_BDD_EXISTS_PRE;
    case CUDD_BDD_AND:
    case CUDD_BDD_XOR:
      pvm_upkint(&n,1,1);
      pvm_upkushort(f_index, n,1);
      pvm_upkushort(f_page,  n,1);
      pvm_upkshort (f_number,n,1);
      pvm_upkushort(g_index, n,1);
      pvm_upkushort(g_page,  n,1);
      pvm_upkshort (g_number,n,1);
      pvm_upkint   (caller,  n,1);
      
      for(i=0;i<n;i++)
	{
	  f.index=f_index[i];
	  f.address.page=f_page[i];
	  f.address.number=f_number[i];
	  g.index=g_index[i];
	  g.address.page=g_page[i];
	  g.address.number=g_number[i];
	  index=ddMin(f.index,g.index);

	  if(cuddLocal(manager,index))
	    DdWorkQueuePush2(manager,op,index,from,f,g,caller[i]);
	  else
	    {
	      /* NOTE cache lookup ? */
	      DdMessageSend2(manager,op,index,from,f,g,caller[i]);
#ifdef DD_COMM_MONITOR
	      pi->Forwards++;
#endif
	    }
	}
      break;
	  
      /*
	case CUDD_BDD_ANDABSTRACT:
	op=CUDD_BDD_ANDABSTRACT_PRE;
	case CUDD_BDD_ITE:
	pvm_upkint(&index,1,1);
	pvm_upkint(&from,1,1);
	PVM_upkDdNodePar(f);
	PVM_upkDdNodePar(g);
	PVM_upkDdNodePar(h);
	PVM_upkDdCallTreeNode(caller);
	if(cuddLocal(manager,index))
	DdWorkQueuePush(manager,op,index,from,f,g,h,caller);
	else
	{ */
      /* NOTE cache lookup ? */
	  /*	      
	    DdMessageSend3(manager,op,index,from,f,g,h,caller);
	    #ifdef DD_COMM_MONITOR
	    pi->Forwards++;
	    #endif
	    }
	    break;
	    */

    case CUDD_FORWARD_OPERATION:
      pvm_upkint(&new_from,1,1);
      DdReadHandleOperation(manager,new_from);
      break;
    case CUDD_FORWARD_REPLY:
      pvm_upkint(&to,1,1);
      pvm_upkint(&n,1,1);
      pvm_upkushort(f_index, n,1);
      pvm_upkushort(f_page,  n,1);
      pvm_upkshort (f_number,n,1);
      pvm_upkint   (caller,  n,1);
      
      for(i=0;i<n;i++)
	{
	  f.index=f_index[i];
	  f.address.page=f_page[i];
	  f.address.number=f_number[i];
	  DdMessageReply(manager,to,f,caller[i]);
	}
      break;
  
    case CUDD_NEW_VAR:
      pvm_upkint(&to,1,1);
      pvm_upkint(&index,1,1);
      assert(index==manager->size);
      if(index>=manager->maxSize) ddResizeTablePar(manager,index); 
      if(to==manager->Mypos)
	{
	  cuddSetLocal(manager,index);
	  f=cuddUniqueInterPar(manager,index,manager->one,
			       Cudd_NotPar(manager->one));
	  pvm_setsbuf(pi->DownBuf);
	  PVM_pkDdNodePar(f);
	  pvm_send(pi->Parent,DD_UP_MESSAGE);
	  pvm_freebuf(pi->DownBuf);
	  pi->DownBuf=pvm_mkbuf(manager->ProcessInfo->Encoding);
	}
      else
	cuddSetOwner(manager,index,to);
      manager->size++;
	  
      if(manager->Mypos!=pi->Nprocess-1)
	{ /* pass info down */
	  pvm_setsbuf(pi->DownBuf);
	  pvm_pkint(&op,1,1);
	  pvm_pkint(&to,1,1);
	  pvm_pkint(&index,1,1);
	  pvm_send(pi->Tids[manager->Mypos+1],DD_DOWN_MESSAGE);
	  pvm_freebuf(pi->DownBuf);
	  pi->DownBuf=pvm_mkbuf(manager->ProcessInfo->Encoding);
	}
      return(0);
      break;
    case CUDD_SEND_INDEX:
      pvm_upkint(&index,1,1);
      pvm_upkint(&to,1,1);
      
      /* to accomodate future options where the index ma have moved
	 with the process making the request knowing it */
      if(!cuddLocal(manager,index))
	DdRetrieveIndex(manager,index,DdFindOwner(manager,index));
      
      DdSendIndex(manager,index,to);
      
      return(0);
      break;
      
    case CUDD_RECIEVE_INDEX:
      pvm_upkint(&index,1,1);
      DdRecieveIndex(manager,index);
      return(0);
      break;
      
    case CUDD_REDISTRIBUTE_LEVELS:
      cuddRedistributeLevelsSlave(manager);
      return(0);
      break;
    case CUDD_RETRIEVE_INDEX:
      pvm_upkint(&index,1,1);
      DdRetrieveIndex(manager,index,DdFindOwner(manager,index));
      return(0);
      break;
      
    case CUDD_REORDER:
      cuddReorder(manager);
      return(0);
      break;

    case CUDD_DUMP_DOT:
      pvm_upkstr(file_name);
      cuddDumpDotSlave(manager,file_name);
      return(0);
      break;
      
    case CUDD_NEW_PROCESS:
      assert(!(int)"No New Process yet");
      break;
    case CUDD_FORCE_REPLY:
      if(pi->UpBuffer.n==0 
	 && manager->WorkQueue.StackPointer==0 
	 && pi->down_flag==TRUE)
	/* processor has no work or available results, so must force
	   replies from processors further down. The down_flag is needed for
	   the special case, where the force happens to cross a reply massege */
	DdForceReply(manager);
      DdMessageSendUp(manager);
      return(0);
      break;
    case CUDD_FLUSH_DOWN:
      pvm_upkint(&flush_count,1,1);
      assert(flush_count>manager->ProcessInfo->flush_count);
      manager->ProcessInfo->flush_count=flush_count;
      manager->ProcessInfo->flush=TRUE;
      return(0);
      break;
    case CUDD_FLUSH_UP:
      pvm_upkint(&flush_count,1,1);
      assert(flush_count==manager->ProcessInfo->flush_count);
      pvm_upkint(manager->allocatedPerLevel,manager->size,1);
      manager->ProcessInfo->flush=TRUE;
      return(0);
      break;
    case CUDD_NO_MORE_MESSAGES:
      return(0);
      break;
    default:
      assert(0); /* the above cases should be exhaustive */
    }
  return(1);
}

/* for now no error handeling here */
/* assume this is only called when processor have no other work */
int DdMessagesRecieve(DdManager *manager)
{
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  if(!pvm_nrecv(-1,DD_UP_MESSAGE))
    { /* upward messages have higher priority */
      pvm_recv(-1,-1); 
      /* if no upward messages immidiatly available
	 do a blocking recieve */
    }  
  while(DdReadHandleOperation(manager,manager->Mypos-1));

  return(1); /* NO ERROR HANDELING */
}


/* for now no error handeling here */
int DdReplyRecieve(DdManager *manager)
{
  if(pvm_nrecv(-1,DD_UP_MESSAGE))
    {
      while(DdReadHandleOperation(manager,manager->Mypos-1));
      /* could be optimised to use only possible upward messages */
    }      
  return(1);
}

int DdForceReply(DdManager *manager)
{
  int op=CUDD_FORCE_REPLY;
  message_buffer *buf;
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  pvm_setsbuf(pi->DownBuf);

  buf=&pi->DownBuffer;
  DdPackDownBuffer(buf);
  pvm_pkint(&op,1,1);
  pvm_send(pi->Tids[manager->Mypos+1],DD_DOWN_MESSAGE);

  pvm_freebuf(pi->DownBuf);
  pi->DownBuf=pvm_mkbuf(pi->Encoding);

  buf->n=0;
  buf->n_op=0;
  pi->DownCountSend*=DD_DOWN_BUFFER_GROWTH;
  if(pi->DownCountSend>DD_DOWN_BUFFER_MAX)
    pi->DownCountSend=DD_DOWN_BUFFER_MAX;
#ifdef DD_COMM_MONITOR
  pi->DownMessages++;
#endif

  pvm_recv(-1,DD_UP_MESSAGE);

  while(DdReadHandleOperation(manager,manager->Mypos-1));
  /* could be optimised to use only possible upward messages */
}


int DdInitSlave(DdManager *manager,int *cache_size, int *cache_size_max)
{
  int i;
  ProcessInfo *pi;
  pi = (ProcessInfo *) ALLOC(ProcessInfo,1);

#ifdef USE_PGTRACE
  pg_tids("y");
#endif

  pi->Mytid = pvm_mytid();
  pi->Parent = pvm_parent();
  /* pvm_setopt(PvmRoute, PvmRouteDirect); */

  pvm_recv(pi->Parent,0);

  pvm_upkint(&pi->Nprocess,1,1);
  pvm_upkint(pi->Tids,pi->Nprocess,1);

  /* find mypos */
  manager->Mypos=0;
  while(pi->Tids[++manager->Mypos]!=pi->Mytid)
    assert(manager->Mypos<pi->Nprocess); 

  pi->flush_count=0;
  pi->flush=FALSE;
  
  pvm_upkint(&pi->Encoding,1,1);

  pi->down_flag=FALSE;
  pi->DownCount=0;
  pi->DownCountSend=DD_DOWN_BUFFER_SIZE;
  pi->DownBuf=pvm_mkbuf(pi->Encoding);

  pi->UpCount=0;
  pi->UpCountSend=DD_UP_BUFFER_SIZE;
  pi->UpBuf=pvm_mkbuf(pi->Encoding);

  pi->DownBuffer.n=0;
  pi->UpBuffer.  n=0;
  pi->DownBuffer.n_op=0;
  pi->UpBuffer.  n_op=1; /* number 0 is always reserved for regular replies */
  pi->UpBuffer.ops[0]=CUDD_BDD_REPLY;
  for(i=0;i<DD_MAX_OPS;i++)
    {
      pi->DownBuffer.n_per_op[i]=0;
      pi->UpBuffer.n_per_op[i]=0;
    }

#ifdef DD_COMM_MONITOR
  pi->DownMessages=0;
  pi->UpMessages=0;
  pi->Forwards=0;
  pi->ITE=0;
  pi->WorkDown=0;
#endif

  PVM_upkDdNodePar(manager->one);
  pvm_upkint(cache_size,1,1);
  pvm_upkint(cache_size_max,1,1);

  manager->ProcessInfo=pi;
  manager->memused += sizeof(ProcessInfo);

  return(1); /* NO ERROR HANDELING */
}

int cuddDumpDotSlave(DdManager *manager,char file_name[])
{
  int op=CUDD_DUMP_DOT;
  int retval;
  FILE *fp;
  ProcessInfo *pi;
  pi=manager->ProcessInfo;

  fp=fopen(file_name,"a");
  if(fp==NULL)
    retval=0;
  else
    {
      retval = cuddDumpDotSlave2(manager,fp);
      fclose(fp);
    }
  if(retval==0 || manager->Mypos==pi->Nprocess-1)
    { /* inform parent of succes/failure */
      pvm_setsbuf(pi->UpBuf);
      pvm_pkint(&retval,1,1);      
      pvm_send(pi->Parent,0);
      pvm_freebuf(pi->UpBuf);
      pi->UpBuf=pvm_mkbuf(pi->Encoding);
    }
  else
    { 
      pvm_setsbuf(pi->DownBuf);
      pvm_pkint(&op,1,1);      
      pvm_pkstr(file_name);
      pvm_send(pi->Tids[manager->Mypos+1],0);
      pvm_freebuf(pi->DownBuf);
      pi->DownBuf=pvm_mkbuf(pi->Encoding);
    }
  return(retval);
}

int DdMessageReply(DdManager *manager,int to,DdNodePar res,DdCallTreeNode caller)
{
  message_buffer *buf;
  int i,op_number;
  
  buf=&manager->ProcessInfo->UpBuffer;

  if(to==manager->Mypos-1)
    {
      op_number=0;      
    }
  else
    {
      op_number=1;
      while(op_number<buf->n_op 
	    && buf->ForwardInfo[op_number].to!=to )
	op_number++;

      if(op_number==DD_MAX_OPS)
	{ /* ran out of space for new operations */
	  DdMessageSendUp(manager);
	  op_number=1;
	}
      if(op_number==buf->n_op)
	{
	  buf->ops[op_number]=CUDD_FORWARD_REPLY;
	  buf->ForwardInfo[op_number].to=to;
	  buf->n_op++;
	}
      else
	assert(buf->ops[op_number]==CUDD_FORWARD_REPLY);
    }




  i=buf->n_per_op[op_number];
  buf->f_index [op_number][i]=res.index;
  buf->f_page  [op_number][i]=res.address.page;
  buf->f_number[op_number][i]=res.address.number;
  buf->caller  [op_number][i]=caller;

  buf->n++;
  buf->n_per_op[op_number]++;

  if(buf->n>=manager->ProcessInfo->UpCountSend)
    DdMessageSendUp(manager);
  return(1);
}

int flush_buffer_up(DdManager *manager)
{
  int type;
  int i,alloc;
  ProcessInfo *pi;
  message_buffer *buf;
  
  pi=manager->ProcessInfo;
  buf=&pi->UpBuffer;

  pvm_setsbuf(pi->UpBuf);

  DdPackUpBuffer(buf);
 
  type=CUDD_FLUSH_UP;
  pvm_pkint(&type,1,1);
  pvm_pkint(&pi->flush_count,1,1);

  for(i=0;i<manager->size;i++)
    if(cuddLocal(manager,i))
      {
	alloc=manager->memory[i].allocated;
	pvm_pkint(&alloc,1,1);
      }
    else
      pvm_pkint(&manager->allocatedPerLevel[i],1,1);

  pvm_send(pi->Tids[manager->Mypos-1],DD_UP_MESSAGE);
  pvm_freebuf(pi->UpBuf);
  pi->UpBuf=pvm_mkbuf(pi->Encoding);
  buf->n=0;
  buf->n_op=1; /* number 0 is always reserved for regular replies */
  pi->UpCountSend=DD_UP_BUFFER_SIZE;
#ifdef DD_COMM_MONITOR
  pi->UpMessages++;
#endif
  return(1); /* NO ERROR HANDELING */
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

static int DdRetrieveIndex(DdManager *manager, int index,int from)
{
  int bufid;
  int op=CUDD_SEND_INDEX;
  int dummy_index;

  bufid = pvm_mkbuf(manager->ProcessInfo->Encoding);
  pvm_setsbuf(bufid);

  pvm_pkint(&op,1,1);
  pvm_pkint(&index,1,1);
  pvm_pkint(&manager->Mypos,1,1);

  pvm_send(manager->ProcessInfo->Tids[from],0);
  pvm_freebuf(bufid);

  pvm_recv(manager->ProcessInfo->Tids[from],0);
  pvm_upkint(&op,1,1);
  assert(op==CUDD_RECIEVE_INDEX);
  pvm_upkint(&dummy_index,1,1);
  assert(dummy_index==index);
  DdRecieveIndex(manager,index);
}

static int DdRecieveIndex(DdManager *manager, int index)
{
  DdNodeAdr *nodelist;
  DdMemory *mem;
  DdNodeOld *page;
  DdNodeMore *more_page;
  int page_number;
  int number,maxnumber;
  int pos;
  DdHalfWord size;
  int i;
  int slots;
 
  mem=&manager->memory[index];
  
  PVM_upkDdHalfWord(&size,1,1); /* not needed, just send to check */

  if(manager->reordering)
    assert(size==mem->TableSize);
  else
    mem->TableSize=size;

  PVM_upkDdHalfWord(&mem->nextPage,1,1);
  PVM_upkDdNodeAdr(mem->nextFree);
  pvm_upkint(&slots,1,1);

  mem->PageTable=ALLOC(DdNodeOld *,mem->TableSize);

  nodelist=ALLOC(DdNodeAdr,slots);
  
  FREE(manager->subtables[index].nodelist);
  manager->subtables[index].nodelist=nodelist;
  manager->subtables[index].slots=slots;
  manager->subtables[index].maxKeys= slots * DD_MAX_SUBTABLE_DENSITY;

  for(i=0;i<slots;i++)
    nodelist[i]=NULL_ADR;

  mem->allocated=0;

  maxnumber=DD_MEM_CHUNK; 
  /* this is only for the case whewre the level is empty */

  for(page_number=0;page_number<mem->nextPage;page_number++) {
    page=mem->PageTable[page_number]=ALLOC(DdNodeOld,DD_MEM_CHUNK)-1;
 
    if(page_number<mem->nextPage-1 || cuddEqualAdr(mem->nextFree,NULL_ADR))
      maxnumber=DD_MEM_CHUNK;
    else
      /* last page, not full */
      maxnumber=mem->nextFree.number-1;

    for(number=1;number<=maxnumber;number++)
      {
	PVM_upkDdNodePar(page[number].type.kids.T);
	PVM_upkDdNodePar(page[number].type.kids.E);
	pos=ddHashPar(page[number].type.kids.T,page[number].type.kids.E,slots);
	page[number].next=nodelist[pos];
	nodelist[pos].page=page_number;
	nodelist[pos].number=number;
      }
    mem->allocated+=maxnumber;
    
    if(manager->reordering)
      {
	if(mem->PageTable2[page_number]==NULL)
	  mem->PageTable2[page_number]=ALLOC(DdNodeMore,DD_MEM_CHUNK)-1;
	  
	more_page=mem->PageTable2[page_number];
	for(number=1;number<=maxnumber;number++)
	  {
	    /* reference count will be reset anyway later, no need to set it */
	    more_page[number].Reorder.Index=index;
	    more_page[number].Reorder.NextIndex=index;
	  }
      }
  }

  manager->allocated+=mem->allocated;

  /* init the free list */

  if(maxnumber==DD_MEM_CHUNK)
    mem->nextFree=NULL_ADR;
  else {
    page_number=mem->nextPage-1;
    mem->nextFree.page=page_number;
    mem->nextFree.number=maxnumber+1;
    for(number=maxnumber+1;number<DD_MEM_CHUNK;number++)
      {
	page[number].next.page=page_number;
	page[number].next.number=number+1;
      }
    page[DD_MEM_CHUNK].next=NULL_ADR;
  }
  manager->subtables[index].keys=mem->allocated;
  cuddSetLocal(manager,index);
  return(1);
}

static int DdSendIndex(DdManager *manager, int index,int to)
{
  DdMemory *mem;
  DdNodeOld *page;
  int page_number,number,maxnumber;
  int bufid;
  int op=CUDD_RECIEVE_INDEX;
  
  mem=&manager->memory[index];
 
  bufid = pvm_mkbuf(manager->ProcessInfo->Encoding);
  pvm_setsbuf(bufid);

  pvm_pkint(&op,1,1);
  pvm_pkint(&index,1,1);

  PVM_pkDdHalfWord(&mem->TableSize,1,1); /* not needed, just send to check */
  PVM_pkDdHalfWord(&mem->nextPage,1,1);
  PVM_pkDdNodeAdr(mem->nextFree);
  pvm_pkint(&manager->subtables[index].slots,1,1);

  for(page_number=0;page_number<mem->nextPage;page_number++) {
    page=mem->PageTable[page_number];
 
    if(page_number<mem->nextPage-1 || cuddEqualAdr(mem->nextFree,NULL_ADR))
      maxnumber=DD_MEM_CHUNK;
    else
      /* last page, not full */
      maxnumber=mem->nextFree.number-1;

    for(number=1;number<=maxnumber;number++)
      {
	PVM_pkDdNodePar(page[number].type.kids.T);
	PVM_pkDdNodePar(page[number].type.kids.E);
      }
    free(page+1);
  }
  pvm_send(manager->ProcessInfo->Tids[to],0);
  

  if(manager->reordering)
    for(page_number=mem->nextPage;page_number<mem->TableSize;page_number++)
      {
	mem->PageTable2[page_number]=NULL;
	/* for non-local indexes pages not in use are expected to be NULL */
      }

  cuddSetOwner(manager,index,to);
  pvm_freebuf(bufid);

  FREE(manager->subtables[index].nodelist);
  FREE(mem->PageTable);
  manager->allocated-=mem->allocated;
  
  return(1); /* NO ERROR HANDELING */
}

static int DdMessageSendUp(DdManager *manager)
{
  ProcessInfo *pi;
  message_buffer *buf;
  int stop_op=CUDD_NO_MORE_MESSAGES;
  
  pi=manager->ProcessInfo;
  buf=&pi->UpBuffer;

  pvm_setsbuf(pi->UpBuf);

  DdPackUpBuffer(buf);

  pvm_pkint(&stop_op,1,1);
  
  pvm_send(pi->Tids[manager->Mypos-1],DD_UP_MESSAGE);

  pvm_freebuf(pi->UpBuf);
  pi->UpBuf=pvm_mkbuf(pi->Encoding);
  
  buf->n=0;
  buf->n_op=1; /* number 0 is always reserved for regular replies */
  
  pi->UpCountSend*=DD_UP_BUFFER_GROWTH;
  if(pi->UpCountSend>DD_UP_BUFFER_MAX)
    pi->UpCountSend=DD_UP_BUFFER_MAX;
#ifdef DD_COMM_MONITOR
  pi->UpMessages++;
#endif
  
  return(1); /* NO ERROR HANDELING */
}
