#include "cuddInt.h"
#include "message.h"

#define PrintNode(node) \
  fprintf(fp,"\"%d,%d,%d\";\n",node.index,node.address.page,node.address.number);

#define PrintNodeElse(node) \
  fprintf(fp,"\"%d,%d,%d\" [style=%s];\n",node.index,node.address.page,abs(node.address.number),node.address.number>0 ? "dashed" : "dotted");

int cuddDumpDotSlave2(DdManager *manager,FILE *fp)
{
  int index,i,max,page_number;
  DdMemory *mem;
  DdNodeOld *page;
  int retval;

  for (index = 0; index < manager->size; index++) 
    if (cuddLocal(manager,index)) {
      /* Write rank info: All nodes with the same index have the same rank. */
      retval = fprintf(fp,"{ rank = same; \" %d \"; ", index);
      if (retval == EOF) return(0);
      mem=&manager->memory[index];
      
      for(page_number=0;page_number<mem->nextPage;page_number++) { 
	if(page_number==mem->nextPage-1)
	  { /* last page */
	    if(cuddEqualAdr(mem->nextFree,NULL_ADR))
	      max=DD_MEM_CHUNK;
	    else
	      max=mem->nextFree.number-1;
	  }
	else
	  max=DD_MEM_CHUNK;
	
	page=mem->PageTable[page_number];
	for(i=1;i<=max;i++) {
	  retval = fprintf(fp,"\"%d,%d,%d\" [label=\"%d,%d\"];",index,page_number,i,page_number,i);
	  if (retval == EOF) return(0);
	}
      }
      retval = fprintf(fp,"}\n");
      if (retval == EOF) return(0);      

      /* write edges */
      for(page_number=0;page_number<mem->nextPage;page_number++) { 
	if(page_number==mem->nextPage-1)
	  { /* last page */
	    if(cuddEqualAdr(mem->nextFree,NULL_ADR))
	      max=DD_MEM_CHUNK;
	    else
	      max=mem->nextFree.number-1;
	  }
	else
	  max=DD_MEM_CHUNK;
	
	page=mem->PageTable[page_number];
	for(i=1;i<=max;i++) {
	  retval = fprintf(fp,"\"%d,%d,%d\" -> ",index,page_number,i);
	  if (retval == EOF) return(0);
	  retval = PrintNode(page[i].type.kids.T);
	  if (retval == EOF) return(0);
	  retval = fprintf(fp,"\"%d,%d,%d\" -> ",index,page_number,i);
	  if (retval == EOF) return(0);
	  retval = PrintNodeElse(page[i].type.kids.E);
	  if (retval == EOF) return(0);
	}
      }
    }
  return(1);
}
