#include "cuddInt.h"
#include "message.h"

extern int ITE[1000];
extern int Prof[1000];
int cur_index;

static void Profile(int index1,int index2)
{
  int i;
  for(i=index2;i>index1;i--)
    {
      Prof[i]++;
    }
}

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddBddAbs.c,v 1.11 1997/04/13 17:29:14 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static DdNodePar cuddBddExistsSimple(DdManager *manager,DdNodePar f, DdNodePar *cube, DdCacheHandle *handle);
static int bddCheckPositiveCube ARGS((DdManager *manager, DdNodePar cube));

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/*  */
void
cuddBddIteRecurPar(DdManager *manager,
		   int index,
		   DdNodePar  f,
		   DdNodePar  g,
		   DdNodePar  h,
		   int from,
		   DdCallTreeNode caller)
{
  DdNodePar fv, fnv, gv, gnv, hv, hnv;
  DdCallTreeNode then_child;
  int op=CUDD_BDD_ITE;
#ifdef DD_COMM_MONITOR
  manager->ProcessInfo->ITE++;
#endif
  
  /* compute cofactors */
  if (f.index==index) {
    fv = Cudd_T_Par(manager,f);
    fnv = Cudd_E_Par(manager,f);
    if (Cudd_IsComplementPar(f)) {
      fv = Cudd_NotPar(fv);
      fnv = Cudd_NotPar(fnv);
    }
  } else {
    fv = fnv = f;
  }

  if (g.index==index) {
    gv = Cudd_T_Par(manager,g);
    gnv = Cudd_E_Par(manager,g);
    if (Cudd_IsComplementPar(g)) {
      gv = Cudd_NotPar(gv);
      gnv = Cudd_NotPar(gnv);
    }
  } else {
    gv = gnv = g;
  }

  if (h.index==index) {
    hv = Cudd_T_Par(manager,h);
    hnv = Cudd_E_Par(manager,h);
    if (Cudd_IsComplementPar(h)) {
      hv = Cudd_NotPar(hv);
      hnv = Cudd_NotPar(hnv);
    }
  } else {
    hv = hnv = h;
  }
    
  then_child=DdCallTreeCreateNodeNormal(manager,caller,from,index);

  cuddBddRecurProcess3(manager,op,fnv,gnv,hnv,DdCallTreeNodeOther(then_child));
  cuddBddRecurProcess3(manager,op,fv,gv,hv,then_child);
  /* doing else before then because work is stored on a stack and by pushing
     else first, you get the same execution order as the sequential version */

}

void 
cuddBddAndAbstractRecurPre(DdManager *manager,
			   DdNodePar  f,
			   DdNodePar  g,
			   DdNodePar  cube,
			   int from,
			   DdCallTreeNode caller)
{
  int op=CUDD_BDD_ANDABSTRACT;
  int index;
  DdCallTreeNode Node;

  index=ddMin(f.index,g.index);

  assert(from!=manager->Mypos); /* should no be call internally */
  Node=DdCallTreeCreateNode_Propagate(manager,caller,from);
  cuddBddRecurProcess3(manager,op,f,g,cube,Node);
}

/*  */
void
cuddBddAndAbstractRecur(DdManager *manager,
			int index,
			DdNodePar  f,
			DdNodePar  g,
			DdNodePar  cube,
			int from,
			DdCallTreeNode caller)
{
  DdNodePar fv, fnv, gv, gnv;
  DdCallTreeNode then_child;
  int op=CUDD_BDD_ANDABSTRACT;

#ifdef DD_COMM_MONITOR
  manager->ProcessInfo->ITE++;
#endif
  
  /* compute cofactors */
  if (f.index==index) {
    fv = Cudd_T_Par(manager,f);
    fnv = Cudd_E_Par(manager,f);
    if (Cudd_IsComplementPar(f)) {
      fv = Cudd_NotPar(fv);
      fnv = Cudd_NotPar(fnv);
    }
  } else {
    fv = fnv = f;
  }

  if (g.index==index) {
    gv = Cudd_T_Par(manager,g);
    gnv = Cudd_E_Par(manager,g);
    if (Cudd_IsComplementPar(g)) {
      gv = Cudd_NotPar(gv);
      gnv = Cudd_NotPar(gnv);
    }
  } else {
    gv = gnv = g;
  }

  if (index == cube.index)
    {
      then_child=DdCallTreeCreateNodeOr(manager,caller,from);
      cube=Cudd_T_Par(manager,cube);
    }
  else  /* if (index < cube.index) */
    then_child=DdCallTreeCreateNodeNormal(manager,caller,from,index);

  cuddBddRecurProcess3(manager,op,fnv,gnv,cube,DdCallTreeNodeOther(then_child));
  cuddBddRecurProcess3(manager,op,fv,gv,cube,then_child);
  /* doing else before then because work is stored on a stack and by pushing
     else first, you get the same execution order as the sequential version */

}

/* generalized recursion procedure for AND and XOR */
void
cuddBddRecur2(DdManager *manager,
	      int index,
	      DdNodePar  f,
	      DdNodePar  g,
	      int from,
	      DdCallTreeNode caller,
	      int op,
	      DdNodePar (*base)(DdManager *,DdNodePar,DdNodePar))
{
  DdNodePar fv, fnv, gv, gnv;
  DdCallTreeNode then_child;

#ifdef DD_COMM_MONITOR
  manager->ProcessInfo->ITE++;
#endif

  assert(ddMin(f.index,g.index) == index);
  
  ITE[index]++;


  /* compute cofactors */
  if (f.index==index) {
    fv = Cudd_T_Par(manager,f);
    fnv = Cudd_E_Par(manager,f);
    if (Cudd_IsComplementPar(f)) {
      fv = Cudd_NotPar(fv);
      fnv = Cudd_NotPar(fnv);
    }
  } else {
    fv = fnv = f;
  }
	
  if (g.index==index) {
    gv = Cudd_T_Par(manager,g);
    gnv = Cudd_E_Par(manager,g);
    if (Cudd_IsComplementPar(g)) {
      gv = Cudd_NotPar(gv);
      gnv = Cudd_NotPar(gnv);
    }
  } else {
    gv = gnv = g;
  }

  assert(fv.address.page<10000);
  assert(fnv.address.page<10000);
  assert(gv.address.page<10000);
  assert(gnv.address.page<10000);

    
  then_child=DdCallTreeCreateNodeNormal(manager,caller,from,index);
  
  cur_index=index;
  cuddBddRecurProcess2(manager,op,fnv,gnv,DdCallTreeNodeOther(then_child));
  cuddBddRecurProcess2(manager,op,fv,gv,then_child);
  /* doing else before then because work is stored on a stack and by pushing
     else first, you get the same execution order as the sequential version */

}

/* Abstract a variable in cube that does not appear in f */
DdNodePar cuddCubeRemoveVars(DdManager *manager,int index,DdNodePar cube)
{
  if(!cuddLocal(manager,index))
    { /* remove all the local variables from cube and pass on the work */
      while(cube.index!=CUDD_MAXINDEX && cuddLocal(manager,cube.index))
	cube=Cudd_T_Par(manager,cube);
    }
  else 
    {
      while(index>cube.index)
	cube=Cudd_T_Par(manager,cube);
    }
  return cube;
}


void cuddBddRecurProcess2(DdManager *manager,int op, DdNodePar f, DdNodePar g,DdCallTreeNode caller)
{
  int index;
  DdNodePar res;
  DdCacheHandle handle;
  switch(op)
    {
    case CUDD_BDD_AND:
      res=cuddBddAndBasePar(manager,f,g);
      break;
    case CUDD_BDD_XOR:
      res=cuddBddXorBasePar(manager,f,g);
      break;
    case CUDD_BDD_EXISTS:
      if(cuddIsConstantPar(f))
	res=f;
      else
	{
	  g=cuddCubeRemoveVars(manager,f.index,g);
	  if(cuddEqualPar(g,DD_ONE_PAR(manager)))
	    res=f;
	  else
	    res=CUDD_ERROR;
	}
      break;
    }

  if(!cuddEqualPar(res,CUDD_ERROR))
    {
      DdCallTreeUpdateNode(manager,caller,res);
    }
  else
    {
      handle = cuddCacheCheck2Par(manager, op, f, g);
      switch(cuddCacheStatusPar(manager,handle))
	{
	case CUDD_CACHE_HIT:
	  DdCallTreeUpdateNode(manager,caller,cuddCacheResult(manager,handle));
	  break;
	case CUDD_CACHE_IN_PROGRESS:
	  cuddCacheWait(manager,handle,caller);
	  break;
	case CUDD_CACHE_MISS:
	  DdCallTreeNodeSetCacheEntry(manager,caller,
				      cuddCacheEntry(manager,handle));
	  index=ddMin(f.index,g.index); /* using index==level */
	  Profile(cur_index,index);
	  if(cuddLocal(manager,index))
	    DdWorkQueuePush2(manager,op,index,manager->Mypos,f,g,caller);
	  else
	    DdMessageSend2(manager,op,index,manager->Mypos,f,g,caller);      
	  break;
	default:
	  assert(0);
	} 
    }
}

void cuddBddRecurProcess3(DdManager *manager,int op, DdNodePar f, DdNodePar g,DdNodePar h,DdCallTreeNode caller)
{
  int index;
  DdNodePar res;
  DdCacheHandle handle;
  DdNodePar not_g,not_h,one,zero;
  not_h=Cudd_NotPar(h);
  not_g=Cudd_NotPar(g);
  one=DD_ONE_PAR(manager);
  zero=Cudd_NotPar(one);
    

  switch(op)
    {
    case CUDD_BDD_ITE:
      /* two variable cases */
      if (cuddEqualPar(g,one) || cuddEqualPar(f,g)) { 
	/* ITE(F,F,H) = ITE(F,1,H) = F + H */
	res=Cudd_NotPar(cuddBddAndBasePar(manager,Cudd_NotPar(f),
					  Cudd_NotPar(g)));
	/* ignoring further processing as OR 
	   to avoid negation on the way up */
      } else if (cuddEqualPar(h,one) || cuddEqualPar(f,not_h)) { 
        /* ITE(F,G,!F) = ITE(F,G,1) = !F + G */
	res=Cudd_NotPar(cuddBddAndBasePar(manager,f,Cudd_NotPar(g)));
	/* ignoring further processing as OR 
	   to avoid negation on the way up */
      } else if (cuddEqualPar(g,zero) || cuddEqualPar(f,not_g)) {  
        /* ITE(F,!F,H) = ITE(F,0,H) = !F * H */	
	cuddBddRecurProcess2(manager,CUDD_BDD_AND,Cudd_NotPar(f),h,caller);
	return;
      } else if (cuddEqualPar(h,zero) || cuddEqualPar(f,h)) {    
	/* ITE(F,G,F) = ITE(F,G,0) = F * G */
	cuddBddRecurProcess2(manager,CUDD_BDD_AND,f,g,caller);
	return;
      } else if (cuddEqualPar(g,not_h)) {
	/* ITE(F,G,!G) = F XOR G */
	cuddBddRecurProcess2(manager,CUDD_BDD_XOR,f,g,caller);
	return;
      }
      /* one variable casees */
      else if (cuddEqualPar(f,one))
	res=g;
      else if (cuddEqualPar(f,zero))
	res=h;
      else if (cuddEqualPar(g,h))
	res=g; 
      else
	res=CUDD_ERROR;
      break;
    case CUDD_BDD_ANDABSTRACT:
      if (cuddEqualPar(f,zero) || cuddEqualPar(g,zero) 
	  || cuddEqualPar(f,not_g) )
	res=zero;
      else if (cuddEqualPar(f,one) && cuddEqualPar(g,one))
	res=one;
      else if (cuddEqualPar(f,one) || cuddEqualPar(f,g)) {
	cuddBddRecurProcess2(manager, CUDD_BDD_EXISTS, g, h,caller);
	return;
      } else if (cuddEqualPar(g,one)) {
	cuddBddRecurProcess2(manager, CUDD_BDD_EXISTS, f, h,caller);
	return;
      } else {
	h=cuddCubeRemoveVars(manager,ddMin(f.index,g.index),h);
	if(cuddEqualPar(h,one))
	  {
	    cuddBddRecurProcess2(manager, CUDD_BDD_AND, f, g,caller);
	    return;
	  }
	else
	  res=CUDD_ERROR;
      }
    }
  if(!cuddEqualPar(res,CUDD_ERROR))
    {
      DdCallTreeUpdateNode(manager,caller,res);
    }
  else
    {
      handle = cuddCacheCheck3Par(manager, op, f, g, h);
      switch(cuddCacheStatusPar(manager,handle))
	{
	case CUDD_CACHE_HIT:
	  DdCallTreeUpdateNode(manager,caller,cuddCacheResult(manager,handle));
	  break;
	case CUDD_CACHE_IN_PROGRESS:
	  cuddCacheWait(manager,handle,caller);
	  break;
	case CUDD_CACHE_MISS:
	  DdCallTreeNodeSetCacheEntry(manager,caller,
				      cuddCacheEntry(manager,handle));
	  index=ddMin(f.index,g.index); /* using index==level */
	  if(cuddLocal(manager,index))
	    DdWorkQueuePush(manager,op,index,manager->Mypos,f,g,h,caller);
	  else
	    DdMessageSend3(manager,op,index,manager->Mypos,f,g,h,caller);      
	  break;
	default:
	  assert(0);
	} 
    }
}


/**Function********************************************************************

  Synopsis    [Performs the recursive steps of Cudd_bddExistAbstract.]

  Description [Performs the recursive steps of Cudd_bddExistAbstract.
  Returns the BDD obtained by abstracting the variables
  of cube from f if successful; NULL otherwise. It is also used by
  Cudd_bddUnivAbstract.]

  SideEffects [None]

  SeeAlso     [Cudd_bddExistAbstract Cudd_bddUnivAbstract]

******************************************************************************/

void 
cuddBddExistAbstractRecurPre(DdManager *manager,
			     DdNodePar  f,
			     DdNodePar  cube,
			     int from,
			     DdCallTreeNode caller)
{
  int op=CUDD_BDD_EXISTS;
  DdCallTreeNode Node;

  assert(from!=manager->Mypos); /* should no be call internally */
  Node=DdCallTreeCreateNode_Propagate(manager,caller,from);
  cuddBddRecurProcess2(manager,op,f,cube,Node);
}
      
void cuddBddExistAbstractRecur(DdManager *manager,
			       DdNodePar  f,
			       DdNodePar  cube,
			       int from,
			       DdCallTreeNode caller)
{
  DdNodePar T,E;
  DdCallTreeNode then_child;
  int op=CUDD_BDD_EXISTS;

  /* assumptions: f not constant, cube not constant, f.index<=cube.index */ 

  /* Compute the cofactors of f. */
  T = Cudd_T_Par(manager,f); 
  E = Cudd_E_Par(manager,f); 
  if(Cudd_IsComplementPar(f)) {
    T=Cudd_NotPar(T);
    E=Cudd_NotPar(E);
  }

  if (f.index == cube.index)
    then_child=DdCallTreeCreateNodeOr(manager,caller,from);
  else  /* if (f.index < cube.index) */
    then_child=DdCallTreeCreateNodeNormal(manager,caller,from,f.index);

  cuddBddRecurProcess2(manager,op,E,cube,DdCallTreeNodeOther(then_child));
  cuddBddRecurProcess2(manager,op,T,cube,then_child);
  
} /* end of cuddBddExistAbstractRecur */

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

static DdNodePar cuddBddExistsSimple(DdManager *manager,DdNodePar f, DdNodePar *cube, DdCacheHandle *handle)
{

  if(cuddIsConstantPar(f))
    {
      return f;
    }

  /* Abstract a variables that does not appear in f */
  if(!cuddLocal(manager,f.index))
    { /* remove all the local variables from cube and pass on the work */
      while(cube->index!=CUDD_MAXINDEX && cuddLocal(manager,cube->index))
	*cube=Cudd_T_Par(manager,*cube);
    }
  else 
    {
      while(f.index>cube->index)
	*cube=Cudd_T_Par(manager,*cube);
    }

  if(cuddEqualPar(*cube,DD_ONE_PAR(manager)))
    /* f does not depend on any of the variables */
    return f;

  *handle = cuddCacheCheck2Par(manager, CUDD_BDD_EXISTS, f, *cube);
  if(cuddCacheStatusPar(manager,*handle)==CUDD_CACHE_HIT)
    {
      return cuddCacheResult(manager,*handle);
    }

  return CUDD_WAIT;
}

static DdNodePar cuddBddAndAbstractSimple(DdManager *manager,int index,DdNodePar f,DdNodePar g, DdNodePar *cube, DdCacheHandle *handle,int *op)
{

  /* Abstract a variables that does not appear in f or g */  
  if(!cuddLocal(manager,index))
    { /* remove all the local variables from cube and pass on the work */
      while(cube->index!=CUDD_MAXINDEX && cuddLocal(manager,cube->index))
	*cube=Cudd_T_Par(manager,*cube);
    }
  else 
    {
      while(index>cube->index)
	*cube=Cudd_T_Par(manager,*cube);
    }

  if(cuddEqualPar(*cube,DD_ONE_PAR(manager)))
    /* neither f nor g depend on any of the variables */
    {
      *op=CUDD_BDD_AND;
      return(CUDD_ERROR); 
      /* the base cases from and should heve been checked */
    }

  *handle = cuddCacheCheck3Par(manager, CUDD_BDD_ANDABSTRACT, f, g, *cube);
  if(cuddCacheStatusPar(manager,*handle)==CUDD_CACHE_HIT)
    {
      return cuddCacheResult(manager,*handle);
    }

  return CUDD_WAIT;
}

/**Function********************************************************************

  Synopsis [Checks whether cube is an BDD representing the product of
  positive literals.]

  Description [Returns 1 in case of success; 0 otherwise.]

  SideEffects [None]

******************************************************************************/
static int
bddCheckPositiveCube(manager, cube)
DdManager *manager;
DdNodePar cube;
{
    DdNodePar zero;
    if (Cudd_IsComplementPar(cube)) return(0);
    if (cuddEqualPar(cube,DD_ONE_PAR(manager))) return(1);
    if (cuddIsConstantPar(cube)) return(0);
    zero=Cudd_NotPar(DD_ONE_PAR(manager));
    if (cuddEqualPar(Cudd_E_Par(manager,cube),zero)) {
        return(bddCheckPositiveCube(manager, Cudd_T_Par(manager,cube)));
    }
    return(0);

} /* end of bddCheckPositiveCube */








